/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

body {
  padding-top: 70px;
}

div.container {
  width: 98%;
  padding-left: 15px;
  padding-right: 15px;
}

a.navbar-brand span {
  color: #fff;
}

.modal-backdrop {
  z-index: 2;
  position: fixed;
}

.modal-content {
  z-index: 5;
}

nav {
  box-shadow: 0 3px 3px #aaa;
  z-index: 999;
}

a.navbar-brand {
  cursor: default;
}

td > span.glyphicon {
  padding-left: 3px;
  padding-top: 3px;
}

button.btn,
label.btn {
  border: 1px solid #000;
}

.btn-group label.btn {
  background-color: #f0f0f0;
}

div.rich_doc {
  padding: 5px 10px;
  border: 1px solid #ddd;
  background: #fff;
  border-radius: 4px;
}

span.status_square {
  width: 10px;
  height: 10px;
  border: 1px solid grey;
  display: inline-block;
  padding-left: 0;
  cursor: pointer;
}

div.squares {
  float: right;
  font-size: 1;
}

span.success {
  background-color: green;
}

span.up_for_retry {
  background-color: gold;
}

span.up_for_reschedule {
  background-color: turquoise;
}

span.started {
  background-color: lime;
}

span.error {
  background-color: red;
}

span.queued {
  background-color: gray;
}

span.upstream_failed {
  background-color: orange;
}

span.skipped {
  background-color: pink;
}

.tooltip-inner {
  max-width: 500px;
  text-align: left !important;
  font-size: 12px;
}

.d3-tip {
  background: #000;
  color: #fff;
  border: solid;
  border-width: 1px;
  border-radius: 5px;
  padding: 10px;
}

input#execution_date {
  margin-bottom: 0;
}

table.highlighttable {
  width: 100%;
  table-layout: fixed;
}

div.linenodiv {
  padding-right: 1px !important;
}

.linenos {
  width: 50px;
  border: none;
}

div.linenodiv pre {
  padding-left: 4px;
  padding-right: 4px;
  color: #aaa;
  background-color: #fcfcfc;
  text-align: right;
}

pre {
  overflow: auto;
  word-wrap: normal;
  white-space: pre;
}

pre code {
  overflow-wrap: normal;
  white-space: pre;
}

.wrap {
  white-space: pre-wrap;
}

input,
select {
  margin: 0;
}

.code {
  font-family: monospace;
}

#sql {
  border: 1px solid #ccc;
  border-radius: 5px;
}

.ace_editor div {
  font: inherit !important;
}

#ace_container {
  margin: 10px 0;
}

#sql_ace {
  visibility: hidden;
}

table.dataframe {
  font-size: 12px;
}

table.dataframe tbody tr td {
  padding: 2px;
}

table thead th {
  background-color: #f3f3f3;
}

table.dataframe.dataTable thead > tr > th {
  padding: 10px 20px 10px 10px;
}

table.dataTable.dataframe thead .sorting {
  background: url('./../sort_both.png') no-repeat center right;
}

table.dataTable.dataframe thead .sorting_desc {
  background: url('./../sort_desc.png') no-repeat center right;
}

table.dataTable.dataframe thead .sorting_asc {
  background: url('./../sort_asc.png') no-repeat center right;
}

.no-wrap {
  white-space: nowrap;
}

div.form-inline {
  margin-bottom: 5px;
}

body div.panel {
  padding: 0;
}

.blur {
  filter: url(#blur-effect-1);
}

div.legend_item {
  border-radius: 5px;
  float: right;
  margin: 0 3px 0 0;
  padding: 0 3px;
  border: solid 2px grey;
  font-size: 11px;
}

div.legend_circle {
  border-radius: 10px;
  width: 15px;
  height: 15px;
  border: 1px solid grey;
  float: left;
  margin-top: 2px;
  margin-right: 0;
}

div.square {
  width: 12px;
  height: 12px;
  float: right;
  margin-top: 2px;
  border: 1px solid #000;
}

.btn:active,
.btn.active {
  box-shadow: inset 0 6px 6px rgba(0, 0, 0, 0.4);
}

label[for="timezone-other"],
#timezone-other {
  margin: 3px 20px;
  font-weight: normal;
}

#timezone-menu ul.typeahead.dropdown-menu {
  max-height: 200px;
  overflow-y: auto;
  overflow-x: hidden;
}

/* depending on the version of FAB in use, we may have a style conflict */
.bootstrap-datetimepicker-widget .datepicker > div {
  display: block;
}

.dag-doc {
  margin-bottom: 15px;
}

.hll { background-color: #ffc; }
.c { color: #408080; font-style: italic; } /* Comment */
.err { border: 1px solid #f00; } /* Error */
.k { color: #008000; font-weight: bold; } /* Keyword */
.o { color: #666 } /* Operator */
.cm { color: #408080; font-style: italic; } /* Comment.Multiline */
.cp { color: #bc7a00; } /* Comment.Preproc */
.c1 { color: #408080; font-style: italic; } /* Comment.Single */
.cs { color: #408080; font-style: italic; } /* Comment.Special */
.gd { color: #a00000; } /* Generic.Deleted */
.ge { font-style: italic; } /* Generic.Emph */
.gr { color: #f00; } /* Generic.Error */
.gh { color: #000080; font-weight: bold; } /* Generic.Heading */
.gi { color: #00a000; } /* Generic.Inserted */
.go { color: #888; } /* Generic.Output */
.gp { color: #000080; font-weight: bold; } /* Generic.Prompt */
.gs { font-weight: bold; } /* Generic.Strong */
.gu { color: #800080; font-weight: bold; } /* Generic.Subheading */
.gt { color: #0044dd; } /* Generic.Traceback */
.kc { color: #008000; font-weight: bold; } /* Keyword.Constant */
.kd { color: #008000; font-weight: bold; } /* Keyword.Declaration */
.kn { color: #008000; font-weight: bold; } /* Keyword.Namespace */
.kp { color: #008000; } /* Keyword.Pseudo */
.kr { color: #008000; font-weight: bold; } /* Keyword.Reserved */
.kt { color: #b00040; } /* Keyword.Type */
.m { color: #666; } /* Literal.Number */
.s { color: #ba2121; } /* Literal.String */
.na { color: #7d9029; } /* Name.Attribute */
.nb { color: #008000; } /* Name.Builtin */
.nc { color: #00f; font-weight: bold; } /* Name.Class */
.no { color: #800; } /* Name.Constant */
.nd { color: #a2f; } /* Name.Decorator */
.ni { color: #999999; font-weight: bold; } /* Name.Entity */
.ne { color: #d2413a; font-weight: bold; } /* Name.Exception */
.nf { color: #00f; } /* Name.Function */
.nl { color: #a0a000; } /* Name.Label */
.nn { color: #00f; font-weight: bold; } /* Name.Namespace */
.nt { color: #008000; font-weight: bold; } /* Name.Tag */
.nv { color: #19177c; } /* Name.Variable */
.ow { color: #a2f; font-weight: bold; } /* Operator.Word */
.w { color: #bbb; } /* Text.Whitespace */
.mb { color: #666 } /* Literal.Number.Bin */
.mf { color: #666 } /* Literal.Number.Float */
.mh { color: #666 } /* Literal.Number.Hex */
.mi { color: #666 } /* Literal.Number.Integer */
.mo { color: #666 } /* Literal.Number.Oct */
.sb { color: #ba2121; } /* Literal.String.Backtick */
.sc { color: #ba2121; } /* Literal.String.Char */
.sd { color: #ba2121; font-style: italic; } /* Literal.String.Doc */
.s2 { color: #ba2121; } /* Literal.String.Double */
.s1 { color: #ba2121; } /* Literal.String.Single */
